@@ -1,6 +1,6 @@
 from __future__ import annotations
 from functools import lru_cache
-from typing import List, Set, Optional
+from typing import List, Set, Optional, Iterator
 from dataclasses import dataclass, field
 import marshmallow as m
 from marshmallow import post_load
@@ -44,6 +44,8 @@ class TextractBlockTypes(Enum):
     KEY_VALUE_SET = auto()
     PAGE = auto()
     SELECTION_ELEMENT = auto()
+    QUERY = auto()
+    QUERY_RESULT = auto()
 
 
 @dataclass
@@ -52,16 +54,20 @@ class TextractEntityTypes(Enum):
     VALUE = auto()
 
 
-@dataclass(init=True, eq=True, repr=True)
+@dataclass(eq=True, repr=True)
 class TPoint():
     x: float
     y: float
 
-    def scale(self, doc_width=None, doc_height=None):
+    def __init__(self, x: float, y: float) -> None:
+        self.x = x
+        self.y = y
+
+    def scale(self, doc_width, doc_height):
         self.x: float = self.x * doc_width
         self.y: float = self.y * doc_height
 
-    def ratio(self, doc_width=None, doc_height=None):
+    def ratio(self, doc_width, doc_height):
         self.x: float = self.x / doc_width
         self.y: float = self.y / doc_height
 
@@ -92,20 +98,26 @@ class TPoint():
         return self
 
 
-@dataclass(eq=True, init=True, repr=True, order=True, unsafe_hash=True)
+@dataclass(eq=True, repr=True, order=True, unsafe_hash=True)
 class TBoundingBox():
     width: float
     height: float
     left: float
     top: float
 
-    def scale(self, doc_width=None, doc_height=None):
+    def __init__(self, height: float, width: float, left: float, top: float) -> None:
+        self.width = width
+        self.height = height
+        self.left = left
+        self.top = top
+
+    def scale(self, doc_width, doc_height):
         self.top: float = self.top * doc_height
         self.height: float = self.height * doc_height
         self.left: float = self.left * doc_width
         self.width: float = self.width * doc_width
 
-    def ratio(self, doc_width=None, doc_height=None):
+    def ratio(self, doc_width, doc_height):
         self.top: float = self.top / doc_height
         self.height: float = self.height / doc_height
         self.left: float = self.left / doc_width
@@ -180,7 +192,7 @@ class TGeometry():
         self.bounding_box.ratio(doc_width=doc_width, doc_height=doc_height)
         [x.ratio(doc_width=doc_width, doc_height=doc_height) for x in self.polygon]
 
-    def rotate(self, origin: TPoint = TPoint(0, 0), degrees: float = 180):
+    def rotate(self, origin: TPoint = TPoint(0, 0), degrees: float = 180.0):
         self.bounding_box.rotate(origin=origin, degrees=degrees)
         [p.rotate(origin_x=origin.x, origin_y=origin.y) for p in self.polygon]
 
@@ -198,6 +210,21 @@ class TGeometrySchema(BaseSchema):
         return TGeometry(**data)
 
 
+@dataclass(eq=True, init=True, repr=True)
+class TQuery:
+    text: str = field(default=None)    # type: ignore
+    alias: str = field(default=None)    # type: ignore
+
+
+class TQuerySchema(BaseSchema):
+    text = m.fields.String(data_key="Text", required=False)
+    alias = m.fields.String(data_key="Alias", required=False)
+
+    @post_load
+    def make_tquery(self, data, **kwargs):
+        return TQuery(**data)
+
+
 @dataclass(eq=True, init=True, repr=True)
 class TRelationship():
     type: str = field(default=None)    #type: ignore
@@ -234,6 +261,7 @@ class TBlock():
     selection_status: str = field(default=None)    #type: ignore
     text_type: str = field(default=None)    #type: ignore
     custom: dict = field(default=None)    #type: ignore
+    query: TQuery = field(default=None)    #type: ignore
 
     def __eq__(self, o: object) -> bool:
         if isinstance(o, TBlock):
@@ -266,7 +294,7 @@ class TBlock():
                 self.relationships = list()
             self.relationships.append(TRelationship(type=relationships_type, ids=ids))
 
-    def rotate(self, origin=TPoint(0.5, 0.5), degrees=180):
+    def rotate(self, origin=TPoint(0.5, 0.5), degrees: float = 180):
         self.geometry.rotate(origin=origin, degrees=degrees)
 
 
@@ -286,6 +314,7 @@ class TBlockSchema(BaseSchema):
     selection_status = m.fields.String(data_key="SelectionStatus", required=False, allow_none=False)
     text_type = m.fields.String(data_key="TextType", required=False, allow_none=False)
     custom = m.fields.Dict(data_key="Custom", required=False, allow_none=False)
+    query = m.fields.Nested(TQuerySchema, data_key="Query")
 
     @post_load
     def make_tblock(self, data, **kwargs):
@@ -397,7 +426,8 @@ class TDocument():
         self.add_block(tblock, page=page_block)
         return tblock
 
-    def add_virtual_key_for_existing_key(self, key_name: str, existing_key: TBlock, page_block: TBlock):
+    def add_virtual_key_for_existing_key(self, key_name: str, existing_key: TBlock,
+                                         page_block: TBlock) -> Optional[TBlock]:
         if existing_key and existing_key.block_type == "KEY_VALUE_SET" and "KEY" in existing_key.entity_types:
             value_blocks: List[TBlock] = self.value_for_key(existing_key)
             return self.add_key_values(key_name=key_name, values=value_blocks, page_block=page_block)
@@ -405,7 +435,7 @@ class TDocument():
             logger.warning(
                 f"no existing_key or not block_type='KEY_VALUE_SET' or 'KEY' not in entity_type: {existing_key}")
 
-    def add_key_values(self, key_name: str, values: List[TBlock], page_block: TBlock):
+    def add_key_values(self, key_name: str, values: List[TBlock], page_block: TBlock) -> TBlock:
         if not key_name:
             raise ValueError("need values and key_name")
         if not values:
@@ -425,27 +455,27 @@ class TDocument():
 
         virtual_block = self.add_virtual_block(text=key_name, page_block=page_block)
         id = str(uuid4())
-        key_block = TBlock(
-            id=id,
-            block_type="KEY_VALUE_SET",
-            entity_types=["KEY"],
-            confidence=99,
-            geometry=TGeometry(bounding_box=TBoundingBox(width=0, height=0, left=0, top=0),
-                               polygon=[TPoint(x=0, y=0), TPoint(x=0, y=0)]),
-        )
+        key_block = TBlock(id=id,
+                           block_type="KEY_VALUE_SET",
+                           entity_types=["KEY"],
+                           confidence=99,
+                           geometry=TGeometry(bounding_box=TBoundingBox(width=0, height=0, left=0, top=0),
+                                              polygon=[TPoint(x=0, y=0), TPoint(x=0, y=0)]),
+                           page=page_block.page)
         key_block.add_ids_to_relationships(relationships_type="VALUE", ids=[value_block.id])
         key_block.add_ids_to_relationships(relationships_type="CHILD", ids=[virtual_block.id])
         logger.debug(f"add key with id: {id} and key_name: {key_name}")
         self.add_block(key_block, page=page_block)
+        return key_block
 
-    def rotate(self, page: TBlock = None, origin: TPoint = TPoint(x=0.5, y=0.5), degrees: float = None) -> None:
+    def rotate(self, page: TBlock, degrees: float, origin: TPoint = TPoint(x=0.5, y=0.5)) -> None:
         # FIXME: add dimension. the relative scale messes up the new coordinates, have to use the actual image scale
         """atm no way to get back from Block to list of other blocks, hence get_block_by_id is only available on document level and quite some processing has to be here"""
         if not page:
             raise ValueError("need a page to rotate")
         if not degrees:
             raise ValueError("need degrees to rotate")
-        [b.rotate(origin=origin, degrees=degrees) for b in self.relationships_recursive(block=page)]
+        [b.rotate(origin=origin, degrees=float(degrees)) for b in self.relationships_recursive(block=page)]
         self.relationships_recursive.cache_clear()
 
     def find_block_by_id(self, id: str) -> Optional[TBlock]:
@@ -460,7 +490,7 @@ class TDocument():
                 return b
         raise ValueError(f"no block for id: {id}")
 
-    def __relationships_recursive(self, block: TBlock) -> List[TBlock]:
+    def __relationships_recursive(self, block: TBlock) -> Iterator[TBlock]:
         import itertools
         if block and block.relationships:
             all_relations = list(itertools.chain(*[r.ids for r in block.relationships if r and r.ids]))
@@ -525,9 +555,32 @@ class TDocument():
         return self.get_blocks_by_type(page=page, block_type_enum=TextractBlockTypes.KEY_VALUE_SET)
 
     def keys(self, page: TBlock = None) -> List[TBlock]:
-        for key_entities in self.forms(page=page):
-            if TextractEntityTypes.KEY.name in key_entities.entity_types:
-                yield key_entities
+        return [x for x in self.forms(page=page) if TextractEntityTypes.KEY.name in x.entity_types]
+        # for key_entities in self.forms(page=page):
+        #     if TextractEntityTypes.KEY.name in key_entities.entity_types:
+        #         yield key_entities
+
+    def queries(self, page: TBlock) -> List[TBlock]:
+        return self.get_blocks_by_type(page=page, block_type_enum=TextractBlockTypes.QUERY)
+
+    def get_answers_for_query(self, block: TBlock) -> List[TBlock]:
+        result_list: List[TBlock] = list()
+        rels = block.get_relationships_for_type(relationship_type="ANSWER")
+        if rels:
+            for r in rels.ids:
+                result_list.append(self.get_block_by_id(r))
+        return result_list
+
+    def get_query_answers(self, page: TBlock) -> List[List[str]]:
+        result_list: List[List[str]] = list()
+        for query in self.queries(page=page):
+            answers = [x for x in self.get_answers_for_query(block=query)]
+            if answers:
+                for answer in answers:
+                    result_list.append([query.query.text, query.query.alias, answer.text])
+            else:
+                result_list.append([query.query.text, query.query.alias, ""])
+        return result_list
 
     def get_key_by_name(self, key_name: str) -> List[TBlock]:
         result_blocks: List[TBlock] = list()
